import apache_beam as beam
import logging


class TimeProcessing(beam.DoFn):

    def to_runner_api_parameter(self, unused_context):
        return "beam:transforms:custom_parsing:custom_v0", None

    def process(self, element, timestamp=beam.DoFn.TimestampParam, window=beam.DoFn.WindowParam):
        from datetime import datetime
        import google.cloud.firestore as firestore
        from txp_cloud.common import utils as cu

        firestore_db = firestore.Client()
        mode = cu.get_signal_mode_from_firestore(element["configuration_id"], element["edge_logical_id"],
                                                 element["perception_name"], firestore_db)
        if mode is None or not cu.SignalMode.is_time(mode):
            return

        logging.info(f"""Storing in time table:
            Edge: {element["edge_logical_id"]}
            Perception: {element["perception_name"]}
            Part index: {element["part_index"]}
            Previous part index: {element["previous_part_index"]}
            Signal Timestamp: {datetime.fromtimestamp(
                            element["signal_timestamp"] // 1000000000).strftime('%Y-%m-%d %H:%M:%S')}
            Package Timestamp: {datetime.fromtimestamp(
                            element["package_timestamp"] // 1000000000).strftime('%Y-%m-%d %H:%M:%S')}
        """)

        yield element


class FftProcessing(beam.DoFn):

    def to_runner_api_parameter(self, unused_context):
        return "beam:transforms:custom_parsing:custom_v0", None

    def process(self, element, timestamp=beam.DoFn.TimestampParam, window=beam.DoFn.WindowParam):
        from datetime import datetime
        from scipy.fft import fft
        import google.cloud.firestore as firestore
        from txp_cloud.common import utils as cu

        firestore_db = firestore.Client()
        mode = cu.get_signal_mode_from_firestore(element["configuration_id"], element["edge_logical_id"],
                                                 element["perception_name"], firestore_db)
        if mode is None or not cu.SignalMode.is_fft(mode):
            return
        data = [list(dimension["values"]) for dimension in element["data"]]
        fft_data = []
        for index, dimension_signal_sample in enumerate(data):
            fft_data.append({
                "values": [],
                "index": index
            })
            n = len(fft_data)
            for z in fft(dimension_signal_sample):
                fft_data[n - 1]["values"].append({"real": float(z.real), "imag": float(z.imag)})

        logging.info(f"""Storing at Fft table:
            Edge: {element["edge_logical_id"]}
            Perception: {element["perception_name"]}
            Part index: {element["part_index"]}
            Previous part index: {element["previous_part_index"]}
            Signal Timestamp: {datetime.fromtimestamp(
                            element["signal_timestamp"] // 1000000000).strftime('%Y-%m-%d %H:%M:%S')}
            Package Timestamp: {datetime.fromtimestamp(
                            element["package_timestamp"] // 1000000000).strftime('%Y-%m-%d %H:%M:%S')}
        """)

        yield {
            "fft": fft_data,
            "package_timestamp": element["package_timestamp"],
            "perception_name": element["perception_name"],
            "edge_logical_id": element["edge_logical_id"],
            "signal_timestamp": element["signal_timestamp"],
            "configuration_id": element["configuration_id"]
        }


class PsdProcessing(beam.DoFn):

    def to_runner_api_parameter(self, unused_context):
        return "beam:transforms:custom_parsing:custom_v0", None

    def process(self, element, timestamp=beam.DoFn.TimestampParam, window=beam.DoFn.WindowParam):
        from datetime import datetime
        import google.cloud.firestore as firestore
        from txp_cloud.common import utils as cu
        from txp_cloud.common import metrics

        firestore_db = firestore.Client()
        mode = cu.get_signal_mode_from_firestore(element["configuration_id"], element["edge_logical_id"],
                                                 element["perception_name"], firestore_db)
        if mode is None or not cu.SignalMode.is_psd(mode):
            return
        data = [list(dimension["values"]) for dimension in element["data"]]
        data_psd = []
        for index, dimension_signal_sample in enumerate(data):
            f, psd = metrics.get_psd(dimension_signal_sample, cu.get_sampling_frequency(element, firestore_db))
            data_psd.append({
                "psd": [float(e) for e in psd],
                "frequency": [float(e) for e in f],
                "index": index,
            })

        logging.info(f"""Storing at Psd table:
            Edge: {element["edge_logical_id"]}
            Perception: {element["perception_name"]}
            Part index: {element["part_index"]}
            Previous part index: {element["previous_part_index"]}
            Signal Timestamp: {datetime.fromtimestamp(
                            element["signal_timestamp"] // 1000000000).strftime('%Y-%m-%d %H:%M:%S')}
            Package Timestamp: {datetime.fromtimestamp(
                            element["package_timestamp"] // 1000000000).strftime('%Y-%m-%d %H:%M:%S')}
        """)

        yield {
            "data": data_psd,
            "package_timestamp": element["package_timestamp"],
            "perception_name": element["perception_name"],
            "edge_logical_id": element["edge_logical_id"],
            "signal_timestamp": element["signal_timestamp"],
            "configuration_id": element["configuration_id"]
        }


class TimeMetrics(beam.DoFn):

    def to_runner_api_parameter(self, unused_context):
        return "beam:transforms:custom_parsing:custom_v0", None

    def process(self, element, timestamp=beam.DoFn.TimestampParam, window=beam.DoFn.WindowParam):
        from txp_cloud.common import metrics

        for dimension in range(0, len(element["data"])):
            yield {
                "edge_logical_id": element["edge_logical_id"],
                "perception_name": element["perception_name"],
                "signal_timestamp": element["signal_timestamp"],
                "package_timestamp": element["package_timestamp"],
                "dimension": dimension,
                "configuration_id": element["configuration_id"],
                "peak": float(metrics.peak(element["data"][dimension]["values"])),
                "rms": float(metrics.rms(element["data"][dimension]["values"])),
                "standard_deviation": float(metrics.standard_deviation(element["data"][dimension]["values"])),
                "crest_factor": float(metrics.crest_factor(element["data"][dimension]["values"]))
            }


class FftMetrics(beam.DoFn):

    def to_runner_api_parameter(self, unused_context):
        return "beam:transforms:custom_parsing:custom_v0", None

    def process(self, element, timestamp=beam.DoFn.TimestampParam, window=beam.DoFn.WindowParam):
        from txp_cloud.common import metrics
        import txp_cloud.common.utils as cu

        data_per_dimension = cu.get_fft_as_np_array(element["fft"])
        for dimension in range(0, len(data_per_dimension)):
            data = data_per_dimension[dimension].real
            yield {
                "edge_logical_id": element["edge_logical_id"],
                "perception_name": element["perception_name"],
                "signal_timestamp": element["signal_timestamp"],
                "package_timestamp": element["package_timestamp"],
                "dimension": dimension,
                "configuration_id": element["configuration_id"],
                "peak": float(metrics.peak(data)),
                "rms": float(metrics.rms(data)),
                "standard_deviation": float(metrics.standard_deviation(data)),
                "crest_factor": float(metrics.crest_factor(data))
            }


class PsdMetrics(beam.DoFn):

    def to_runner_api_parameter(self, unused_context):
        return "beam:transforms:custom_parsing:custom_v0", None

    def process(self, element, timestamp=beam.DoFn.TimestampParam, window=beam.DoFn.WindowParam):
        from txp_cloud.common import metrics
        import numpy as np

        for dimension in range(0, len(element["data"])):
            f = element["data"][dimension]["frequency"]
            psd = element["data"][dimension]["psd"]
            integrated_psd = metrics.integrate(psd, x=f)
            yield {
                "edge_logical_id": element["edge_logical_id"],
                "perception_name": element["perception_name"],
                "signal_timestamp": element["signal_timestamp"],
                "package_timestamp": element["package_timestamp"],
                "dimension": dimension,
                "configuration_id": element["configuration_id"],
                "rms": float(np.sqrt(integrated_psd[-1])),
                "standard_deviation": float(metrics.standard_deviation(psd)),
                "crest_factor": float(metrics.crest_factor(psd)),
                "peak_frequency": float(metrics.peak(f)),
                "peak_amplitude": float(metrics.peak(psd))
            }

